/* xscreensaver-command, Copyright (c) 1991-2003 Jamie Zawinski <jwz@jwz.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 */

/* Great; a "put me in your docs" license.  That means we have to credit jwz
 * for writing xscreensaver--so we can use his code to disable it.  Ironic? */

#include "global.h"

#include <stdio.h>
#include <stdlib.h>

#include <X11/Xproto.h>		/* for CARD32 */
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xos.h>

Atom XA_SCREENSAVER, XA_SCREENSAVER_VERSION, XA_SCREENSAVER_STATUS, XA_BLANK, XA_LOCK, XA_DEACTIVATE;


static XErrorHandler old_handler = 0;
static bool got_badwindow = false;
static int BadWindow_ehandler( Display *dpy, XErrorEvent *error )
{
	if( error->error_code != BadWindow )
	{
		ASSERT( old_handler );
		return (*old_handler) (dpy, error);
	}

	got_badwindow = true;
	return 0;
}

static Window find_screensaver_window( Display *dpy )
{
	Window root = RootWindowOfScreen( DefaultScreenOfDisplay(dpy) );
	Window root2, parent, *kids;
	unsigned int nkids;
	if( !XQueryTree(dpy, root, &root2, &parent, &kids, &nkids) )
		return 0; // unexpected

	if( root != root2 || parent )
		return 0; // unexpected
	if( !kids || !nkids )
		return 0;

	for( unsigned i = 0; i < nkids; ++i )
	{
		/* We're walking the list of root-level windows and trying to find the one that has
		 * a particular property on it.  We need to trap BadWindows errors while doing this,
		 * because it's possible that some random window might get deleted in the meantime.
		 * (That window won't have been the one we're looking for.) */
		XSync( dpy, False );

		ASSERT( !old_handler );
		got_badwindow = false;
		old_handler = XSetErrorHandler (BadWindow_ehandler);
		Atom type;
		int format;
		unsigned long nitems, bytesafter;
		char *v;
		int status = XGetWindowProperty (dpy, kids[i],
				XA_SCREENSAVER_VERSION,
				0, 200, False, XA_STRING,
				&type, &format, &nitems, &bytesafter,
				(unsigned char **) &v);
		XSync( dpy, False );
		XSetErrorHandler( old_handler );
		old_handler = 0;

		if( got_badwindow )
		{
			got_badwindow = false;
			continue;
		}

		XFree( v );
		if( status == Success && type != None )
		{
			Window ret = kids[i];
			XFree( kids );
			return ret;
		}
	}

	if( kids )
		XFree( kids );
	return 0;
}

/* Note that if you kill xscreensaver and restart it, the status won't be updated
 * immediately.  (xscreensaver-command -time will still show "blanked" if you
 * kill it while blanked and restart it.) */
enum ScreenSaverStatus { SCREENSAVER_NOT_SHOWN, SCREENSAVER_SHOWN, SCREENSAVER_SHOWN_AND_LOCKED };
static ScreenSaverStatus GetScreenSaverStatus( Display *dpy )
{
	Atom type;
	int format;
	unsigned long nitems, bytesafter;
	CARD32 *data = 0;

	if( XGetWindowProperty( dpy, RootWindow (dpy, 0), XA_SCREENSAVER_STATUS, 0, 999, False, XA_INTEGER,
				&type, &format, &nitems, &bytesafter, (unsigned char **) &data ) != Success || !type || !data)
	{
		if( data )
			XFree( data );
		// no saver status on root window
		return SCREENSAVER_NOT_SHOWN;
	}
	
	if( type != XA_INTEGER || nitems < 3 )
	{
		if( data )
			free( data );

		// bad status format on root window
		return SCREENSAVER_NOT_SHOWN;
	}

        Atom blanked = (Atom) data[0];
	if( data )
		free( data );

	if( blanked == XA_BLANK )
		return SCREENSAVER_SHOWN;
	if( blanked == XA_LOCK )
		return SCREENSAVER_SHOWN_AND_LOCKED;
	return SCREENSAVER_NOT_SHOWN;
}

void delay_xscreensaver( Display *dpy )
{
	ASSERT( dpy );

	XA_SCREENSAVER = XInternAtom( dpy, "SCREENSAVER", False );
	XA_SCREENSAVER_VERSION = XInternAtom( dpy, "_SCREENSAVER_VERSION",False );
	XA_SCREENSAVER_STATUS = XInternAtom( dpy, "_SCREENSAVER_STATUS", False );
	XA_DEACTIVATE = XInternAtom( dpy, "DEACTIVATE", False );
	XA_BLANK = XInternAtom( dpy, "BLANK", False );
	XA_LOCK = XInternAtom( dpy, "LOCK", False );

	XSync( dpy, 0 );

	if( GetScreenSaverStatus( dpy ) == SCREENSAVER_SHOWN_AND_LOCKED )
		return;

	Window window = find_screensaver_window( dpy );
	if( !window )
		return;

	XEvent event;
	event.xany.type = ClientMessage;
	event.xclient.display = dpy;
	event.xclient.window = window;
	event.xclient.message_type = XA_SCREENSAVER;
	event.xclient.format = 32;
	memset (&event.xclient.data, 0, sizeof(event.xclient.data));
	event.xclient.data.l[0] = (long) XA_DEACTIVATE;
	event.xclient.data.l[1] = 0;
	event.xclient.data.l[2] = 0;
	XSendEvent( dpy, window, False, 0L, &event );

	XSync (dpy, 0);
}

